"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import socket
import urllib.error
import http.client
import http.cookiejar
import json

import pytz

from ...optconf import Option
from ...optconf import SecretOption
from ...optconf.converters import as_string_or_none

from ... import tfile
from ... import web

from .. import BasePlugin
from .. import BaseExtension


# =====
class FetcherError(Exception):
    pass


class AuthError(FetcherError):
    pass


class LogicError(FetcherError):
    pass


class NetworkError(FetcherError):
    def __init__(self, sub):
        super().__init__()
        self._sub = sub

    def __str__(self):
        return "{}: {}".format(type(self._sub).__name__, str(self._sub))


# =====
def _assert(exception, arg, msg=""):
    if not arg:
        raise exception(msg)


class BaseFetcher(BasePlugin):  # pylint: disable=too-many-instance-attributes
    def __init__(self, timeout, retries, retries_sleep, user_agent, client_agent, proxy_url,
                 check_version, check_fingerprint, **_):

        self._timeout = timeout
        self._retries = retries
        self._retries_sleep = retries_sleep
        self._user_agent = user_agent
        self._client_agent = client_agent
        self._proxy_url = proxy_url
        self._check_version = check_version
        self._check_fingerprint = check_fingerprint

        self._comment_regexp = None
        self._retry_codes = (500, 502, 503)
        self._cookie_jar = None
        self._opener = None

    @classmethod
    def get_version(cls):
        raise NotImplementedError

    @classmethod
    def get_fingerprint(cls):
        raise NotImplementedError

    @classmethod
    def get_options(cls):
        return {
            "timeout":           Option(default=10.0, type=float, help="Timeout for HTTP client"),
            "retries":           Option(default=20, help="The number of retries to handle tracker-specific HTTP errors"),
            "retries_sleep":     Option(default=1.0, help="Sleep interval between failed retries"),
            "user_agent":        Option(default="Mozilla/5.0", help="User-agent for site"),
            "client_agent":      Option(default="rtorrent/0.9.2/0.13.2", help="User-agent for tracker"),
            "proxy_url":         Option(default=None, type=as_string_or_none, help="URL of HTTP/SOCKS4/SOCKS5 proxy"),
            "check_fingerprint": Option(default=True, help="Check the site fingerprint"),
            "check_version":     Option(default=True, help="Check the fetcher version from GitHub"),
        }

    def is_torrent_changed(self, torrent):
        raise NotImplementedError

    def fetch_new_data(self, torrent):
        raise NotImplementedError

    # ===

    def _init_opener(self, with_cookies):
        if with_cookies:
            self._cookie_jar = http.cookiejar.CookieJar()
            self._opener = web.build_opener(self._proxy_url, self._cookie_jar)
        else:
            self._opener = web.build_opener(self._proxy_url)

    def _build_opener(self):
        return web.build_opener(self._proxy_url)

    def _read_url(self, *args, **kwargs):
        try:
            return self._read_url_nofe(*args, **kwargs)
        except (
            socket.timeout,
            urllib.error.HTTPError,
            urllib.error.URLError,
            http.client.IncompleteRead,
            http.client.BadStatusLine,
            ConnectionResetError,
        ) as err:
            raise NetworkError(err)

    def _read_url_nofe(self, url, data=None, headers=None, opener=None):
        opener = (opener or self._opener)
        assert opener is not None

        headers = (headers or {})
        headers.setdefault("User-Agent", self._user_agent)

        return web.read_url(
            opener=opener,
            url=url,
            data=data,
            headers=headers,
            timeout=self._timeout,
            retries=self._retries,
            retries_sleep=self._retries_sleep,
            retry_codes=self._retry_codes,
        )

    # ===

    def is_matched_for(self, torrent):
        assert self._comment_regexp is not None
        return (self._comment_regexp.match(torrent.get_comment() or "") is not None)

    def _assert_logic(self, arg, *args):
        _assert(LogicError, arg, *args)

    def _assert_match(self, torrent):
        self._assert_logic(self.is_matched_for(torrent), "No match with torrent")

    def _assert_valid_data(self, data):
        msg = "Received an invalid torrent data: {} ...".format(repr(data[:20]))
        self._assert_logic(tfile.is_valid_data(data), msg)

    # ===

    def test(self):
        if self._check_fingerprint or self._check_version:
            opener = web.build_opener(self._proxy_url)
            info = self._get_upstream_info(opener)
        if self._check_fingerprint:
            self._test_fingerprint(info["fingerprint"], opener)
        if self._check_version:
            self._test_version(info["version"])

    def _get_upstream_info(self, opener):
        try:
            return json.loads(self._read_url_nofe(
                url="https://raw.githubusercontent.com/mdevaev/emonoda/master/fetchers/{}.json".format(self.get_name()),
                opener=opener,
            ).decode("utf-8"))
        except urllib.error.HTTPError as err:
            if err.code == 404:
                return self._get_local_info()
            raise

    @classmethod
    def _get_local_info(cls):
        return {
            "fingerprint": cls.get_fingerprint(),
            "version":     cls.get_version(),
        }

    def _test_fingerprint(self, fingerprint, opener):
        data = self._read_url(fingerprint["url"], opener=opener)
        msg = "Invalid site body, maybe tracker is blocked"
        try:
            page = data.decode(fingerprint["encoding"])
        except UnicodeDecodeError:
            raise FetcherError(msg)
        _assert(FetcherError, fingerprint["text"] in page, msg)

    def _test_version(self, upstream):
        local = self.get_version()
        _assert(FetcherError, local >= upstream, "Fetcher is outdated (ver. local:{}, upstream:{})."
                                                 " I recommend to update the program".format(local, upstream))


class WithLogin(BaseExtension):
    def __init__(self, user, passwd, **_):
        self._user = user
        self._passwd = passwd

    @classmethod
    def get_options(cls):
        return {
            "user":   Option(default=None, type=as_string_or_none, help="Site login"),
            "passwd": SecretOption(default=None, type=as_string_or_none, help="Site password"),
        }

    def login(self):
        raise NotImplementedError

    def _assert_auth(self, *args):
        _assert(AuthError, *args)


class WithCaptcha(BaseExtension):
    def __init__(self, captcha_decoder, **_):
        self._captcha_decoder = captcha_decoder


class WithTime(BaseExtension):
    def __init__(self, timezone, **_):
        self._default_timezone = timezone

    @classmethod
    def get_options(cls):
        return {
            "timezone": Option(
                default=None,
                type=as_string_or_none,
                help="Site timezone, is automatically detected if possible (or manually, 'Europe/Moscow' for example)",
            ),
        }

    def _select_tzinfo(self, site_timezone):
        if site_timezone is None or self._default_timezone is not None:
            return self._get_default_tzinfo()
        try:
            return pytz.timezone(site_timezone)
        except pytz.UnknownTimeZoneError:
            return self._get_default_tzinfo()

    def _get_default_tzinfo(self):
        msg = "Can't determine timezone of site, your must configure it manually"
        self._assert_logic(self._default_timezone is not None, msg)  # pylint: disable=no-member
        return pytz.timezone(self._default_timezone)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import fnmatch

from .. import tfile


# =====
def load_from_dir(path, name_filter, log):
    if not log.isatty():
        log.info("Loading torrents from {cyan}%s/{yellow}%s{reset} ...", (path, name_filter))

    torrents = {}
    for name in sorted(os.listdir(path)):
        if fnmatch.fnmatch(name, name_filter):
            file_path = os.path.abspath(os.path.join(path, name))
            if log.isatty():
                log.info("Loading torrents from {cyan}%s/{yellow}%s{reset} -- {yellow}%s{reset} ...",
                         (path, name_filter, name), one_line=True)
            try:
                torrents[name] = tfile.Torrent(path=file_path)
            except ValueError:
                log.error("Found broken torrent: {cyan}%s/{yellow}%s{reset}", (path, name))
                torrents[name] = None
            except Exception:
                log.error("Can't process torrent: {cyan}%s/{yellow}%s{reset}", (path, name))
                raise

    log.info("Loaded {magenta}%d{reset} torrents from {cyan}%s/{yellow}%s{reset}",
             (len(torrents), path, name_filter))
    return torrents


def by_hash(torrents):
    return {
        torrent.get_hash(): torrent
        for torrent in filter(None, torrents.values())
    }


# =====
def find(path, items, pass_hash):
    return [_find_torrent(path, item, pass_hash) for item in items]


def _find_torrent(path, item, pass_hash):
    if os.path.exists(item):
        return tfile.Torrent(path=os.path.abspath(item))
    if os.path.sep not in item:
        full_path = os.path.join(path, item)
        if os.path.exists(full_path):
            return tfile.Torrent(path=full_path)
    if pass_hash and tfile.is_hash(item.strip()):
        return item.strip()
    raise RuntimeError("Can't find torrent: {}".format(item))



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse
import re

from datetime import datetime

from . import BaseFetcher
from . import WithLogin
from . import WithTime


# =====
def _encode(arg):
    return arg.encode("cp1251")


def _decode(arg):
    return arg.decode("cp1251")


class Plugin(BaseFetcher, WithLogin, WithTime):
    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=(self._user is not None))
        self._comment_regexp = re.compile(r"http://tfile\.(me|ru)/forum/viewtopic\.php\?p=(\d+)")
        self._tzinfo = None

    @classmethod
    def get_name(cls):
        return "tfile.me"

    @classmethod
    def get_version(cls):
        return 0

    @classmethod
    def get_fingerprint(cls):
        return {
            "url":      "http://tfile.me",
            "encoding": "cp1251",
            "text":     "href=\"http://tfile.me/opensearch.xml\"",
        }

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    # ===

    def is_torrent_changed(self, torrent):
        self._assert_match(torrent)
        page = _decode(self._read_url(torrent.get_comment(), opener=self._build_opener()))

        hash_match = re.search(r"<a href=\"magnet:\?xt=urn:btih:([a-fA-F0-9]{40})", page)
        if hash_match is not None:
            torrent_hash = hash_match.group(1).lower()
            return (torrent.get_hash() != torrent_hash)
        else:
            date_match = re.search(r"class=\"regDate\">(\d\d\d\d-\d\d-\d\d \d\d:\d\d)</span><br/>", page)
            if date_match is None:
                page = _decode(self._read_url(torrent.get_comment()))
                date_match = re.search(r"Зарегистрирован: (\d\d\d\d-\d\d-\d\d \d\d:\d\d),", page)
                self._assert_logic(date_match is not None, "Unknown hash and date")
            date = date_match.group(1) + " " + datetime.now(self._tzinfo).strftime("%z")

            upload_time = int(datetime.strptime(date, "%Y-%m-%d %H:%M %z").strftime("%s"))
            return (torrent.get_mtime() <= upload_time)

    def fetch_new_data(self, torrent):
        self._assert_match(torrent)
        page = _decode(self._read_url(torrent.get_comment()))

        dl_id_match = re.search(r"<a href=\"download.php\?id=(\d+)\" class=\"dlLink\"", page)
        self._assert_logic(dl_id_match is not None, "Unknown dl_id")
        dl_id = dl_id_match.group(1)

        data = self._read_url("http://tfile.me/forum/download.php?id={}".format(dl_id))
        self._assert_valid_data(data)
        return data

    # ===

    def login(self):
        if self._user is not None:
            self._assert_auth(self._passwd is not None, "Required user for site")
            self._assert_auth(self._passwd is not None, "Required password for site")
            post = {
                "username":  _encode(self._user),
                "password":  _encode(self._passwd),
                "autologin": b"",
                "login":     b"",
            }
            post_data = _encode(urllib.parse.urlencode(post))
            page = _decode(self._read_url("http://tfile.me/login/", data=post_data))
            self._assert_auth("class=\"nick u\">{}</a>".format(self._user) in page, "Invalid user or password")

            page = _decode(self._read_url("http://tfile.me/forum/viewtopic.php?t=579690"))
            self._tzinfo = self._get_tzinfo(page)
        else:
            self._tzinfo = self._select_tzinfo("Europe/Moscow")

    def _get_tzinfo(self, page):
        timezone_match = re.search(r"<span class=\"gensmall\">Часовой пояс: (GMT [+-] \d{1,2})</span>", page)
        timezone = (timezone_match and "Etc/" + timezone_match.group(1).replace(" ", ""))
        return self._select_tzinfo(timezone)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

from . import nnm_club_me


# =====
class Plugin(nnm_club_me.Plugin):
    _domain = "ipv6.nnm-club.me"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._comment_regexp = re.compile(r"http://ipv6\.nnm-club\.(me|ru)/forum/viewtopic\.php\?p=(\d+)")

    @classmethod
    def get_name(cls):
        return "ipv6.nnm-club.me"

    @classmethod
    def get_version(cls):
        return 0



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse

from ...optconf import Option
from ...optconf.converters import as_string_list

from ... import web

from . import BaseConfetti
from . import WithProxy


# =====
class Plugin(BaseConfetti, WithProxy):
    def __init__(self, api_keys, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)

        self._api_keys = api_keys

    @classmethod
    def get_name(cls):
        return "nma"

    @classmethod
    def get_options(cls):
        return cls._get_merged_options({
            "api_keys": Option(default=["CHANGE_ME"], type=as_string_list, help="API keys (or one key as string)"),
        })

    # ===

    def send_results(self, source, results):
        for result in results["affected"].values():
            self._notify(
                app="Emonoda ({})".format(source),
                event="{}".format(result["torrent"].get_name()),
                description=self._format_description(result["diff"]),
            )

    # ===

    def _format_description(self, diff):
        description_lines = []
        for (sign, items) in (
            ("+", diff["added"]),
            ("-", diff["removed"]),
            ("~", diff["modified"]),
            ("?", diff["type_modified"]),
        ):
            for item in sorted(items):
                description_lines.append("[{}] {}".format(sign, item))
        description = "Affected {} files\n".format(len(description_lines)) + "\n".join(description_lines)
        if len(description) > 9999:
            description = description[:9995] + "..."
        return description

    def _notify(self, app, event, description):
        # http://www.notifymyandroid.com/api.jsp
        post = {
            "apikey":       ",".join(self._api_keys),
            "application":  app,
            "event":        event,
            "description":  description,
            "Content-Type": "text/plain",
        }
        web.read_url(
            opener=web.build_opener(proxy_url=self._proxy_url),
            url="https://www.notifymyandroid.com/publicapi/notify",
            data=urllib.parse.urlencode(post).encode("utf-8"),
            timeout=self._timeout,
            retries=self._retries,
            retries_sleep=self._retries_sleep,
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import itertools
import operator
import argparse
import datetime

from ..plugins.clients import NoSuchTorrentError

from ..helpers import tcollection

from .. import tfile
from .. import fmt

from . import init
from . import get_configured_log
from . import get_configured_client


# =====
def format_size_pretty(torrent):
    return fmt.format_size(torrent.get_size())


def format_announce(torrent):
    return (torrent.get_announce() or "")


def format_announce_list(torrent):
    return list(itertools.chain.from_iterable(torrent.get_announce_list()))


def format_announce_list_pretty(torrent):
    return " ".join(format_announce_list(torrent))


def format_creation_date(torrent):
    return (torrent.get_creation_date() or "")


def format_creation_date_pretty(torrent):
    date = torrent.get_creation_date()
    return str(datetime.datetime.utcfromtimestamp(date) if date is not None else "")


def format_created_by(torrent):
    return (torrent.get_created_by() or "")


def format_provides(torrent):
    return sorted(torrent.get_files())


def format_is_private(torrent):
    return str(int(torrent.is_private()))


def format_is_private_pretty(torrent):
    return ("yes" if torrent.is_private() else "no")


def format_comment(torrent):
    return (torrent.get_comment() or "")


def _format_client_method(torrent, client, method_name):
    assert client is not None, "Required a client"
    try:
        return getattr(client, method_name)(torrent)
    except NoSuchTorrentError:
        return ""


def format_client_path(torrent, client):
    return _format_client_method(torrent, client, "get_full_path")


def format_client_prefix(torrent, client):
    return _format_client_method(torrent, client, "get_data_prefix")


def _make_formatted_tree(files, depth=0, prefix="    "):
    text = ""
    for (name, sub) in sorted(files.items(), key=operator.itemgetter(0)):
        text += prefix + "*   " * depth + name + "\n"
        text += _make_formatted_tree(sub, depth + 1)
    return text


def format_files_tree(torrent):
    tree = {}
    for path in torrent.get_files():
        parts = os.path.normpath(path).split(os.path.sep)
        local = tree
        for (index, part) in enumerate(parts):
            if index != 0 and len(part) == 0:
                continue
            local.setdefault(part, {})
            local = local[part]
    return _make_formatted_tree(tree)


def print_pretty_meta(torrent, client, log):
    log.print("{blue}Path:{reset}          %s", (torrent.get_path(),))
    log.print("{blue}Name:{reset}          %s", (torrent.get_name(),))
    log.print("{blue}Hash:{reset}          %s", (torrent.get_hash(),))
    log.print("{blue}Size:{reset}          %s", (format_size_pretty(torrent),))
    log.print("{blue}Announce:{reset}      %s", (format_announce(torrent),))
    log.print("{blue}Announce list:{reset} %s", (format_announce_list_pretty(torrent),))
    log.print("{blue}Creation date:{reset} %s", (format_creation_date_pretty(torrent),))
    log.print("{blue}Created by:{reset}    %s", (format_created_by(torrent),))
    log.print("{blue}Private:{reset}       %s", (format_is_private_pretty(torrent),))
    log.print("{blue}Comment:{reset}       %s", (format_comment(torrent),))
    if client is not None:
        log.print("{blue}Client path:{reset}   %s", (format_client_path(torrent, client),))
    if torrent.is_single_file():
        log.print("{blue}Provides:{reset}      %s", (tuple(torrent.get_files())[0],))
    else:
        log.print("{blue}Provides:{reset}\n%s", (format_files_tree(torrent),))


# ===== Main =====
def main():  # pylint: disable=too-many-locals
    options = None  # Makes pylint happy
    actions = [
        (option, option[2:].replace("-", "_"), method)
        for (option, method) in (
            ("--path",                 tfile.Torrent.get_path),
            ("--name",                 tfile.Torrent.get_name),
            ("--hash",                 tfile.Torrent.get_hash),
            ("--comment",              format_comment),
            ("--size",                 tfile.Torrent.get_size),
            ("--size-pretty",          format_size_pretty),
            ("--announce",             format_announce),
            ("--announce-list",        format_announce_list),
            ("--announce-list-pretty", format_announce_list_pretty),
            ("--creation-date",        format_creation_date),
            ("--creation-date-pretty", format_creation_date_pretty),
            ("--created-by",           format_created_by),
            ("--provides",             format_provides),
            ("--is-private",           format_is_private),
            ("--is-private-pretty",    format_is_private_pretty),
            ("--client-path",          lambda torrent: format_client_path(torrent, client)),
            ("--client-prefix",        lambda torrent: format_client_prefix(torrent, client)),
            ("--make-magnet",          lambda torrent: torrent.make_magnet(options.magnet_fields)),
        )
    ]

    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emfile",
        description="Show a metadata of torrent file",
        parents=[parent_parser],
    )
    for (option, dest, _) in actions:
        args_parser.add_argument(option, dest=dest, action="store_true")
    args_parser.add_argument("--without-headers", action="store_true")
    args_parser.add_argument("--magnet-fields", nargs="+", default=(), metavar="<fields>", choices=("names", "trackers", "size"))
    args_parser.add_argument("-v", "--verbose", action="store_true")
    args_parser.add_argument("torrents", nargs="+", metavar="<path>")
    options = args_parser.parse_args(argv[1:])

    to_print = [
        (option[2:], method)
        for (option, dest, method) in actions
        if getattr(options, dest)
    ]
    torrents = tcollection.find(config.core.torrents_dir, options.torrents, False)

    with get_configured_log(config, False, sys.stdout) as log_stdout:
        with get_configured_log(config, (not options.verbose), sys.stderr) as log_stderr:
            client = get_configured_client(
                config=config,
                required=False,
                with_customs=False,
                log=log_stderr,
            )

            for torrent in torrents:
                if len(to_print) == 0:
                    print_pretty_meta(torrent, client, log_stdout)
                else:
                    for (header, method) in to_print:
                        line = ("%s" if options.without_headers else "{blue}%s:{reset} %s")
                        retval = method(torrent)
                        if isinstance(retval, (list, tuple)):
                            for item in retval:
                                log_stdout.print(line, (header, str(item)))
                        else:
                            log_stdout.print(line, (header, str(retval)))
                if len(torrents) > 1:
                    log_stdout.print()


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import json
import shutil
import contextlib
import traceback
import threading
import operator
import argparse

from ..plugins.fetchers import FetcherError

from ..helpers import tcollection
from ..helpers import surprise

from .. import tfile
from .. import fmt

from . import init
from . import get_configured_log
from . import get_configured_client
from . import get_configured_fetchers
from . import get_configured_confetti


# =====
class OpContext:
    def __init__(self, feeder, torrent, fetcher):
        self._feeder = feeder
        self.torrent = torrent
        self.fetcher = fetcher

        self._status = None
        self._attrs = {}

    def done_not_in_client(self):
        self._status = "not_in_client"

    def done_affected(self, diff):
        self._status = "affected"
        self._attrs = {"diff": diff}

    def __enter__(self):
        self._feeder._start_op()  # pylint: disable=protected-access

    def __exit__(self, exc_type, exc, tb):
        self._feeder._stop_op()  # pylint: disable=protected-access
        if isinstance(exc, FetcherError):
            self._status = "fetcher_error"
            self._attrs = {
                "err_name": type(exc).__name__,
                "err_msg": str(exc),
            }
        elif isinstance(exc, Exception):
            self._status = "unhandled_error"
            self._attrs = {"tb_lines": traceback.format_exception(exc_type, exc, tb)}
        assert exc_type is None, "\n".join(traceback.format_exception(exc_type, exc, tb))
        if self._status is None:
            self._status = "passed"


class Feeder:  # pylint: disable=too-many-instance-attributes
    def __init__(self, fetchers, torrents, show_unknown, show_passed, show_diff, log_stdout, log_stderr):
        self._fetchers = fetchers
        self._torrents = torrents
        self._show_unknown = show_unknown
        self._show_passed = show_passed
        self._show_diff = show_diff
        self._log_stdout = log_stdout
        self._log_stderr = log_stderr

        self._current_count = 0
        self._current_file_name = None
        self._current_torrent = None
        self._current_fetcher = None

        self._fan = fmt.make_fan()
        self._fan_thread = None
        self._stop_fan = threading.Event()

        self._status_mapping = {
            "invalid":         self._done_invalid,
            "not_in_client":   self._done_not_in_client,
            "unknown":         self._done_unknown,
            "passed":          self._done_passed,
            "affected":        self._done_affected,
            "fetcher_error":   self._done_fetcher_error,
            "unhandled_error": self._done_unhandled_error,
        }

        self._results = {status: {} for status in self._status_mapping}

    def get_ops(self):
        for (self._current_count, (self._current_file_name, self._current_torrent)) in enumerate(
            sorted(self._torrents.items(), key=operator.itemgetter(0))
        ):
            self._current_fetcher = None

            if self._current_torrent is None:
                self._done("invalid", {})
                continue

            self._current_fetcher = self._select_fetcher()
            if self._current_fetcher is None:
                self._done("unknown", {})
                continue

            op = OpContext(self, self._current_torrent, self._current_fetcher)
            yield op
            self._done(op._status, op._attrs)  # pylint: disable=protected-access

        self._log_stdout.finish()

    def get_results(self):
        return self._results

    def _select_fetcher(self):
        for fetcher in self._fetchers:
            if fetcher.is_matched_for(self._current_torrent):
                return fetcher
        return None

    def _done(self, status, attrs):
        self._status_mapping[status](attrs)
        result = {
            "torrent": self._current_torrent,
            "fetcher": self._current_fetcher,
        }
        result.update(attrs)
        self._results[status][self._current_file_name] = result

    def _done_invalid(self, _):
        self._log_stdout.print(*self._format_fail("red", "!", "INVALID_TORRENT"))

    def _done_unknown(self, _):
        one_line = (not self._show_unknown and self._log_stdout.isatty())
        self._log_stdout.print(*self._format_fail("yellow", " ", "UNKNOWN"), one_line=one_line)

    def _done_passed(self, _):
        one_line = (not self._show_passed and self._log_stdout.isatty())
        self._log_stdout.print(*self._format_status("blue", " "), one_line=one_line)

    def _done_affected(self, attrs):
        self._log_stdout.print(*self._format_status("green", "+"))
        if self._show_diff:
            self._log_stdout.print(*fmt.format_torrents_diff(attrs["diff"], "\t"))

    def _done_not_in_client(self, _):
        self._log_stdout.print(*self._format_fail("red", "!", "NOT_IN_CLIENT"))

    def _done_fetcher_error(self, attrs):
        (line, placeholders) = self._format_status("red", "-")
        line += " :: {red}%s({reset}%s{red}){reset}"
        placeholders += (attrs["err_name"], attrs["err_msg"])
        self._log_stdout.print(line, placeholders)

    def _done_unhandled_error(self, attrs):
        self._log_stdout.print(*self._format_status("red", "-"))
        self._log_stdout.print("%s", ("\n".join("\t" + row for row in attrs["tb_lines"]),))

    def _start_op(self):
        if self._log_stdout.isatty():
            def loop():
                while not self._stop_fan.wait(timeout=0.1):
                    self._log_stdout.print(*self._format_status("magenta", next(self._fan)), one_line=True)
            self._fan_thread = threading.Thread(target=loop, daemon=True)
            self._fan_thread.start()
        else:
            self._log_stdout.print(*self._format_status("magenta", " "))

    def _stop_op(self):
        self._stop_fan.set()
        self._fan_thread.join()
        self._stop_fan.clear()

    def _format_fail(self, color, sign, error):
        return (
            "[{" + color + "}%s{reset}] %s {" + color + "}%s {cyan}%s{reset}",
            (sign, self._format_progress(), error, self._current_file_name),
        )

    def _format_status(self, color, sign):
        return (
            "[{" + color + "}%s{reset}] %s {" + color + "}%s {cyan}%s{reset} -- %s",
            (
                sign, self._format_progress(), self._current_fetcher.get_name(),
                self._current_file_name, (self._current_torrent.get_comment() or ""),
            ),
        )

    def _format_progress(self):
        return fmt.format_progress(self._current_count + 1, len(self._torrents))


def backup_torrent(torrent, backup_dir_path, backup_suffix):
    backup_suffix = fmt.format_now(backup_suffix)
    backup_file_path = os.path.join(backup_dir_path, os.path.basename(torrent.get_path()) + backup_suffix)
    shutil.copyfile(torrent.get_path(), backup_file_path)


def make_sub_name(path, prefix, suffix):
    return os.path.join(
        os.path.dirname(path),
        prefix + os.path.basename(path) + suffix,
    )


@contextlib.contextmanager
def client_hooks(client, torrent, to_save_customs, to_set_customs):
    if client is not None:
        prefix = client.get_data_prefix(torrent)
        if len(to_save_customs) != 0:
            customs = client.get_customs(torrent, to_save_customs)
        else:
            customs = {}
        meta_file_path = make_sub_name(torrent.get_path(), ".", ".meta")
        with open(meta_file_path, "w") as meta_file:
            meta_file.write(json.dumps({
                "prefix": prefix,
                "customs": customs,
            }))
        client.remove_torrent(torrent)

    yield  # Torrent-object was changed

    if client is not None:
        client.load_torrent(torrent, prefix)
        customs.update({
            key: fmt.format_now(value)
            for (key, value) in to_set_customs.items()
        })
        if len(customs) != 0:
            client.set_customs(torrent, customs)
        os.remove(meta_file_path)


def update_torrent(client, torrent, new_data, to_save_customs, to_set_customs):
    data_file_path = make_sub_name(torrent.get_path(), ".", ".newdata")
    with open(data_file_path, "wb") as data_file:
        data_file.write(new_data)
    with client_hooks(client, torrent, to_save_customs, to_set_customs):
        os.replace(data_file_path, torrent.get_path())
        torrent.load_from_data(new_data, torrent.get_path())


def update(
    feeder,
    client,
    backup_dir_path,
    backup_suffix,
    to_save_customs,
    to_set_customs,
    noop,
):
    hashes = (client.get_hashes() if client is not None else [])
    for op in feeder.get_ops():
        try:
            with op:
                if op.torrent.get_hash() not in hashes:
                    op.done_not_in_client()
                    continue

                if op.fetcher.is_torrent_changed(op.torrent):
                    new_data = op.fetcher.fetch_new_data(op.torrent)
                    diff = tfile.get_difference(op.torrent, tfile.Torrent(data=new_data))
                    if not noop:
                        if backup_dir_path is not None:
                            backup_torrent(op.torrent, backup_dir_path, backup_suffix)
                        update_torrent(client, op.torrent, new_data, to_save_customs, to_set_customs)
                    op.done_affected(diff)
        except Exception:
            pass


def print_results(results, log):
    for (msg, field) in (
        ("Updated:          %d", "affected"),
        ("Passed:           %d", "passed"),
        ("Not in client:    %d", "not_in_client"),
        ("Unknown:          %d", "unknown"),
        ("Invalid torrents: %d", "invalid"),
        ("Fetcher errors:   %d", "fetcher_error"),
        ("Unhandled errors: %d", "unhandled_error"),
    ):
        log.info(msg, (len(results[field]),))


# ===== Main =====
def main():
    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emfetch",
        description="Update torrent files from popular trackers",
        parents=[parent_parser],
    )
    args_parser.add_argument("-f", "--name-filter", default=None, metavar="<wildcard_pattern>")
    args_parser.add_argument("-y", "--only-fetchers", default=[], nargs="+", metavar="<fetcher>")
    args_parser.add_argument("-x", "--exclude-fetchers", default=[], nargs="+", metavar="<fetcher>")
    args_parser.add_argument("--noop", action="store_true")
    args_parser.add_argument("--mute", action="store_true")
    options = args_parser.parse_args(argv[1:])

    with get_configured_log(config, False, sys.stdout) as log_stdout:
        with get_configured_log(config, False, sys.stderr) as log_stderr:

            client = get_configured_client(
                config=config,
                required=False,
                with_customs=(len(config.emfetch.save_customs) or len(config.emfetch.set_customs)),
                log=log_stderr,
            )

            def read_captcha(url):
                log_stderr.info("{yellow}Enter the captcha{reset} from [{blue}%s{reset}]: ", (url,), no_nl=True)
                return input()

            fetchers = get_configured_fetchers(
                config=config,
                captcha_decoder=read_captcha,
                only=options.only_fetchers,
                exclude=options.exclude_fetchers,
                log=log_stderr,
            )

            if not options.mute:
                confetti = get_configured_confetti(
                    config=config,
                    log=log_stderr,
                )

            torrents = tcollection.load_from_dir(
                path=config.core.torrents_dir,
                name_filter=(options.name_filter or config.emfetch.name_filter),
                log=log_stderr,
            )

            feeder = Feeder(
                fetchers=fetchers,
                torrents=torrents,
                show_unknown=config.emfetch.show_unknown,
                show_passed=config.emfetch.show_passed,
                show_diff=config.emfetch.show_diff,
                log_stdout=log_stdout,
                log_stderr=log_stderr,
            )

            update(
                feeder=feeder,
                client=client,
                backup_dir_path=config.emfetch.backup_dir,
                backup_suffix=config.emfetch.backup_suffix,
                to_save_customs=config.emfetch.save_customs,
                to_set_customs=config.emfetch.set_customs,
                noop=options.noop,
            )

            results = feeder.get_results()
            print_results(
                results=results,
                log=log_stderr,
            )
            if not options.mute:
                if len(results["affected"]) != 0:
                    if not surprise.deploy_surprise(
                        source="emfetch",
                        results=results,
                        confetti=confetti,
                        log=log_stderr,
                    ):
                        raise SystemExit(1)


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import json


# =====
def build_raw_from_options(options):
    raw = {}
    for option in options:
        (key, value) = (option.split("=", 1) + [None])[:2]
        if len(key.strip()) == 0:
            raise ValueError("Empty option key (required 'key=value' instead of '{}')".format(option))
        if value is None:
            raise ValueError("No value for key '{}'".format(key))

        section = raw
        subs = list(map(str.strip, key.split("/")))
        for sub in subs[:-1]:
            section.setdefault(sub, {})
            section = section[sub]
        section[subs[-1]] = _parse_value(value)
    return raw


def _parse_value(value):
    value = value.strip()
    if (
        not value.isdigit()
        and value not in ("true", "false", "null")
        and not value.startswith("{")
        and not value.startswith("[")
        and not value.startswith("\"")
    ):
        value = "\"{}\"".format(value)
    return json.loads(value)


# =====
def make_config(raw, scheme, keys=()):
    if not isinstance(raw, dict):
        raise ValueError("The node '{}' must be a dictionary".format("/".join(keys) or "/"))

    config = Section()
    for (key, option) in scheme.items():
        full_key = keys + (key,)
        full_name = "/".join(full_key)

        if isinstance(option, Option):
            value = raw.get(key, option.default)
            try:
                value = option.type(value)
            except Exception:
                raise ValueError("Invalid value '{value}' for key '{key}'".format(key=full_name, value=value))
            config[key] = value
            config._set_meta(  # pylint: disable=protected-access
                name=key,
                secret=isinstance(option, SecretOption),
                default=option.default,
                help=option.help,
            )
        elif isinstance(option, dict):
            config[key] = make_config(raw.get(key, {}), option, full_key)
        else:
            raise RuntimeError("Incorrect scheme definition for key '{}':"
                               " the value is {}, not dict or [Secret]Option()".format(full_name, type(option)))
    return config


class Section(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self._meta = {}

    def _set_meta(self, name, secret, default, help):  # pylint: disable=redefined-builtin
        self._meta[name] = {
            "secret":  secret,
            "default": default,
            "help":    help,
        }

    def _is_secret(self, name):
        return self._meta[name]["secret"]

    def _get_default(self, name):
        return self._meta[name]["default"]

    def _get_help(self, name):
        return self._meta[name]["help"]

    def __getattribute__(self, name):
        if name in self:
            return self[name]
        else:  # For pickling
            return dict.__getattribute__(self, name)


_type = type


class Option:
    def __init__(self, default, help, type=None):  # pylint: disable=redefined-builtin
        self.default = default
        self.help = help
        self.type = type or (_type(default) if default is not None else str)

    def __repr__(self):
        return "<Option(default={self.default}, type={self.type}, help={self.help})>".format(self=self)


class SecretOption(Option):
    pass



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse
import re

from ... import tfile

from . import BaseFetcher
from . import WithLogin


# =====
def _encode(arg):
    return arg.encode("cp1251")


def _decode(arg):
    return arg.decode("cp1251")


class Plugin(BaseFetcher, WithLogin):
    _domain = "nnm-club.me"

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)
        self._comment_regexp = re.compile(r"http://nnm-club\.(me|ru)/forum/viewtopic\.php\?p=(\d+)")

    @classmethod
    def get_name(cls):
        return "nnm-club.me"

    @classmethod
    def get_version(cls):
        return 1

    @classmethod
    def get_fingerprint(cls):
        return {
            "url":      "http://{}".format(cls._domain),
            "encoding": "cp1251",
            "text":     "<link rel=\"canonical\" href=\"http://nnm-club.me/\">",
        }

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    # ===

    def is_torrent_changed(self, torrent):
        self._assert_match(torrent)
        data = self._read_url(
            url="http://bt.{}:2710/scrape?info_hash={}".format(self._domain, torrent.get_scrape_hash()),
            headers={"User-Agent": self._client_agent},
        )
        return (len(tfile.decode_data(data).get("files", {})) == 0)

    def fetch_new_data(self, torrent):
        self._assert_match(torrent)
        page = _decode(self._read_url(torrent.get_comment().replace("nnm-club.ru", "nnm-club.me")))

        torrent_id_match = re.search(r"filelst.php\?attach_id=([a-zA-Z0-9]+)", page)
        self._assert_logic(torrent_id_match is not None, "Unknown torrent_id")
        torrent_id = torrent_id_match.group(1)

        data = self._read_url("http://{}//forum/download.php?id={}".format(self._domain, torrent_id))
        self._assert_valid_data(data)
        return data

    # ===

    def login(self):
        self._assert_auth(self._user is not None, "Required user for site")
        self._assert_auth(self._passwd is not None, "Required password for site")
        post = {
            "username": _encode(self._user),
            "password": _encode(self._passwd),
            "redirect": b"",
            "login":    b"\xc2\xf5\xee\xe4",
        }
        page = _decode(self._read_url(
            url="http://{}/forum/login.php".format(self._domain),
            data=_encode(urllib.parse.urlencode(post)),
        ))
        self._assert_auth("[ {} ]".format(self._user) in page, "Invalid user or password")



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

from ...optconf import Option

from . import BaseFetcher


# =====
def _decode(arg):
    return arg.decode("utf-8")


class Plugin(BaseFetcher):
    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=False)
        self._comment_regexp = re.compile(r"^http://rutor\.org/torrent/(\d+)$")

    @classmethod
    def get_name(cls):
        return "rutor.org"

    @classmethod
    def get_version(cls):
        return 2

    @classmethod
    def get_fingerprint(cls):
        return {
            "url":      "http://zerkalo-rutor.org",
            "encoding": "utf-8",
            "text":     "<a href=\"/\"><img src=\"/s/logo.jpg\" alt=\"rutor.org logo\" /></a>",
        }

    @classmethod
    def get_options(cls):
        return cls._get_merged_options({
            "user_agent": Option(default="Googlebot/2.1", help="User-agent for site"),
        })

    # ===

    def is_torrent_changed(self, torrent):
        self._assert_match(torrent)
        page = _decode(self._read_url(torrent.get_comment().replace("rutor.org", "zerkalo-rutor.org")))
        hash_match = re.search(r"<div id=\"download\">\s+<a href=\"magnet:\?xt=urn:btih:([a-fA-F0-9]{40})", page)
        self._assert_logic(hash_match is not None, "Hash not found")
        return (torrent.get_hash() != hash_match.group(1).lower())

    def fetch_new_data(self, torrent):
        self._assert_match(torrent)
        topic_id = self._comment_regexp.match(torrent.get_comment()).group(1)
        data = self._read_url("http://zerkalo-rutor.org/download/{}".format(topic_id))
        self._assert_valid_data(data)
        return data



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import argparse

import chardet

from ..helpers import tcollection
from ..helpers import datacache

from .. import fmt

from . import init
from . import get_configured_log
from . import get_configured_client


# =====
def build_used_files(cache, data_roots):
    files = dict.fromkeys(data_roots, None)
    for info in cache["torrents"].values():
        prefix = os.path.normpath(info["prefix"])

        for (path, meta) in info["files"].items():
            files[os.path.join(prefix, path)] = (meta or {}).get("size")

        for data_root_path in data_roots:
            if prefix.startswith(data_root_path):
                parts = list(filter(None, prefix[len(data_root_path):].split(os.path.sep)))
                for index in range(len(parts)):
                    path = os.path.join(*([data_root_path] + parts[:index + 1]))
                    files[path] = None
                break
    return files


def build_all_files(data_root_path):
    files = {}
    for (prefix, _, local_files) in os.walk(data_root_path):
        files[get_decoded_path(prefix)] = None
        for name in local_files:
            path = os.path.join(prefix, name)
            try:
                size = os.path.getsize(path)
            except FileNotFoundError:
                continue
            files[get_decoded_path(path)] = size
    return files


def get_decoded_path(path):
    try:
        path.encode()
        return path
    except UnicodeEncodeError:
        path_bytes = os.fsencode(path)
        try:
            return path_bytes.decode("cp1251")
        except UnicodeDecodeError:
            encoding = chardet.detect(path)["encoding"]
            assert encoding is not None, "Can't determine encoding for bytes string: '{}'".format(repr(path_bytes))
            return path_bytes.decode(encoding)


# =====
def print_orphaned_files(cache, data_roots, dirs_only, log_stdout, log_stderr):
    all_files = {}
    for data_root_path in data_roots:
        log_stderr.info("Scanning directory {cyan}%s{reset} ... ", (data_root_path,))
        all_files.update(build_all_files(data_root_path))

    log_stderr.info("Transposing the cache: by-hashes -> files ...")
    used_files = build_used_files(cache, data_roots)

    files = set(all_files).difference(used_files)
    if len(files) != 0:
        log_stderr.info("Orhpaned files:")
        size = 0
        for path in sorted(files):
            is_dir = (all_files[path] is None)
            size += (all_files[path] or 0)
            line = ("{blue}D" if is_dir else "{magenta}F") + "{reset} %s"
            if dirs_only and is_dir or not dirs_only:
                log_stdout.print(line, (path,))
        log_stderr.info("Found {red}%d{reset} orphaned files = {red}%s{reset}",
                        (len(files), fmt.format_size(size)))
    else:
        log_stderr.info("No orphaned files founded")


def print_not_in_client(client, torrents_dir_path, name_filter, log_stdout, log_stderr):
    torrents = tcollection.load_from_dir(torrents_dir_path, name_filter, log_stderr)
    torrents = tcollection.by_hash(torrents)

    log_stderr.info("Fetching all hashes from client ...")
    client_hashes = client.get_hashes()

    not_in_client = set(torrents).difference(client_hashes)
    if len(not_in_client) != 0:
        log_stderr.info("Not in client:")
        for torrent_hash in not_in_client:
            log_stdout.print("%s", torrents[torrent_hash].get_path())
        log_stderr.info("Found {red}%d{reset} unregistered torrents", (len(not_in_client),))
    else:
        log_stderr.info("No unregistered files founded")


def print_missing_torrents(client, torrents_dir_path, name_filter, log_stdout, log_stderr):
    torrents = tcollection.load_from_dir(torrents_dir_path, name_filter, log_stderr)
    torrents = tcollection.by_hash(torrents)

    log_stderr.info("Fetching all hashes from client ...")
    client_hashes = client.get_hashes()

    missing_torrents = set(client_hashes).difference(torrents)
    if len(missing_torrents) != 0:
        log_stderr.info("Missing torrents for:")
        for torrent_hash in missing_torrents:
            log_stdout.print("%s -- %s", (torrent_hash, client.get_file_name(torrent_hash)))
        log_stderr.info("Found {red}%d{reset} torrents without torrent-files", (len(missing_torrents),))
    else:
        log_stderr.info("No torrents without torrent-files founded")


# ===== Main =====
def main():
    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emfind",
        description="Querying the client",
        parents=[parent_parser],
    )
    commands = args_parser.add_subparsers(dest="cmd")

    commands.add_parser(
        name="rebuild-cache",
        help="Rebuild files cache",
    )

    commands.add_parser(
        name="orphans",
        help="Find files that do not belong to the client",
    ).add_argument("--dirs-only", action="store_true")

    commands.add_parser(
        name="not-in-client",
        help="Find torrent files, which are not registered in the client",
    )

    commands.add_parser(
        name="missing-torrents",
        help="Find torrents registered in the client for which there is no torrent files",
    )

    options = args_parser.parse_args(argv[1:])

    with get_configured_log(config, False, sys.stdout) as log_stdout:
        with get_configured_log(config, False, sys.stderr) as log_stderr:

            def get_client():
                return get_configured_client(
                    config=config,
                    required=True,
                    with_customs=False,
                    log=log_stderr,
                )

            def get_cache(force_rebuild):
                return datacache.get_cache(
                    cache_path=config.emfind.cache_file,
                    force_rebuild=force_rebuild,
                    client=get_client(),
                    torrents_dir_path=config.core.torrents_dir,
                    name_filter=config.emfind.name_filter,
                    log=log_stderr,
                )

            if options.cmd == "rebuild-cache":
                get_cache(True)

            elif options.cmd == "orphans":
                print_orphaned_files(
                    cache=get_cache(False),
                    data_roots=(config.core.data_root_dir,) + tuple(config.core.another_data_root_dirs),
                    dirs_only=options.dirs_only,
                    log_stdout=log_stdout,
                    log_stderr=log_stderr,
                )

            elif options.cmd in ("not-in-client", "missing-torrents"):
                {
                    "not-in-client":    print_not_in_client,
                    "missing-torrents": print_missing_torrents,
                }[options.cmd](
                    client=get_client(),
                    torrents_dir_path=config.core.torrents_dir,
                    name_filter=config.emfind.name_filter,
                    log_stdout=log_stdout,
                    log_stderr=log_stderr,
                )


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse
import http.cookiejar
import re

from . import BaseFetcher
from . import WithLogin


# =====
def _encode(arg):
    return arg.encode("cp1251")


def _decode(arg):
    return arg.decode("cp1251")


class Plugin(BaseFetcher, WithLogin):
    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)
        self._comment_regexp = re.compile(r"http://pravtor\.(ru|spb\.ru)/viewtopic\.php\?p=(\d+)")

    @classmethod
    def get_name(cls):
        return "pravtor.ru"

    @classmethod
    def get_version(cls):
        return 0

    @classmethod
    def get_fingerprint(cls):
        return {
            "url":      "http://pravtor.ru",
            "encoding": "cp1251",
            "text":     "<img src=\"/images/pravtor_beta1.png\"",
        }

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    # ===

    def is_torrent_changed(self, torrent):
        self._assert_match(torrent)
        page = _decode(self._read_url(torrent.get_comment()))
        hash_match = re.search(r"<span id=\"tor-hash\">([a-zA-Z0-9]+)</span>", page)
        self._assert_logic(hash_match is not None, "Hash not found")
        return (torrent.get_hash() != hash_match.group(1).lower())

    def fetch_new_data(self, torrent):
        self._assert_match(torrent)

        topic_id = self._comment_regexp.match(torrent.get_comment()).group(1)

        page = _decode(self._read_url(torrent.get_comment()))
        torrent_id_match = re.search(r"<a href=\"download.php\?id=(\d+)\" class=\"(leech|seed|gen)med\">", page)
        self._assert_logic(torrent_id_match is not None, "Torrent-ID not found")
        torrent_id = int(torrent_id_match.group(1))

        cookie = http.cookiejar.Cookie(
            version=0,
            name="bb_dl",
            value=topic_id,
            port=None,
            port_specified=False,
            domain="",
            domain_specified=False,
            domain_initial_dot=False,
            path="/",
            path_specified=True,
            secure=False,
            expires=None,
            discard=True,
            comment=None,
            comment_url=None,
            rest={"HttpOnly": None},
            rfc2109=False,
        )
        self._cookie_jar.set_cookie(cookie)

        data = self._read_url(
            url="http://pravtor.ru/download.php?id={}".format(torrent_id),
            data=b"",
            headers={
                "Referer": "http://pravtor.ru/viewtopic.php?t={}".format(topic_id),
                "Origin":  "http://pravtor.ru",
            }
        )

        self._assert_valid_data(data)
        return data

    # ===

    def login(self):
        self._assert_auth(self._user is not None, "Required user for site")
        self._assert_auth(self._passwd is not None, "Required passwd for site")
        post = {
            "login_username": _encode(self._user),
            "login_password": _encode(self._passwd),
            "login":          b"\xc2\xf5\xee\xe4",
        }
        page = _decode(self._read_url(
            url="http://pravtor.ru/login.php",
            data=_encode(urllib.parse.urlencode(post)),
        ))
        self._assert_auth(re.search(r"<!--login form-->", page) is None, "Invalid user or password")



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import pkgutil
import textwrap

import mako.template

from ...optconf import Option
from ...optconf.converters import as_string_or_none

from .. import BasePlugin
from .. import BaseExtension


# =====
def templated(name, **kwargs):
    data = pkgutil.get_data(__name__, os.path.join("templates", name))
    template = textwrap.dedent(data.decode()).strip()
    return mako.template.Template(template).render(**kwargs).strip()


# =====
class BaseConfetti(BasePlugin):
    def __init__(self, timeout, retries, retries_sleep, **_):
        self._timeout = timeout
        self._retries = retries
        self._retries_sleep = retries_sleep

    @classmethod
    def get_options(cls):
        return {
            "timeout":       Option(default=10, help="Network timeout"),
            "retries":       Option(default=5, help="Retries for failed attempts"),
            "retries_sleep": Option(default=1, help="Sleep interval between failed attempts"),
        }

    def send_results(self, source, results):
        raise NotImplementedError


class WithProxy(BaseExtension):
    def __init__(self, proxy_url, **_):
        self._proxy_url = proxy_url

    @classmethod
    def get_options(cls):
        return {
            "proxy_url": Option(default=None, type=as_string_or_none, help="URL of HTTP/SOCKS4/SOCKS5 proxy"),
        }



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse
import re

from ... import tfile

from . import BaseFetcher
from . import WithLogin


# =====
def _encode(arg):
    return arg.encode("utf-8")


def _decode(arg):
    return arg.decode("utf-8")


class Plugin(BaseFetcher, WithLogin):
    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)
        self._comment_regexp = re.compile(r"http://tabun\.everypony\.ru/blog/torrents/(\d+)\.html")
        self._last_data = None

    @classmethod
    def get_name(cls):
        return "tabun.everypony.ru"

    @classmethod
    def get_version(cls):
        return 0

    @classmethod
    def get_fingerprint(cls):
        return {
            "url":      "http://tabun.everypony.ru",
            "encoding": "utf-8",
            "text":     "<link href=\"http://tabun.everypony.ru/templates/skin/synio/images/favicon.ico?v1\" "
                        "rel=\"shortcut icon\" />",
        }

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    # ===

    def is_torrent_changed(self, torrent):
        self._last_data = None
        self._assert_match(torrent)
        candidate = self._get_candidate(torrent)
        if torrent.get_hash() != candidate.get_hash():
            self._last_data = candidate.get_data()
            return True
        return False

    def _get_candidate(self, torrent):
        download_id = self._comment_regexp.match(torrent.get_comment()).group(1)
        data = self._read_url("http://tabun.everypony.ru/file/go/{}/".format(download_id))
        self._assert_valid_data(data)
        candidate = tfile.Torrent(data=data)
        return candidate

    def fetch_new_data(self, torrent):
        return self._last_data

    # ===

    def login(self):
        self._assert_auth(self._user is not None, "Required user for site")
        self._assert_auth(self._passwd is not None, "Required passwd for site")

        page = _decode(self._read_url("http://tabun.everypony.ru"))
        key_match = re.search(r"var LIVESTREET_SECURITY_KEY = '([0-9a-fA-F]+)';", page)
        self._assert_auth(key_match is not None, "Unknown LIVESTREET_SECURITY_KEY")
        key = key_match.group(1)

        post = {
            "login":           _encode(self._user),
            "password":        _encode(self._passwd),
            "remember":        b"on",
            "security_ls_key": _encode(key),
        }
        post_data = _encode(urllib.parse.urlencode(post))
        page = _decode(self._read_url(
            url="http://tabun.everypony.ru/login/ajax-login/",
            data=post_data,
            headers={"Referer": "http://tabun.everypony.ru/"},
        ))
        self._assert_auth("\"bStateError\":false" in page, "Invalid user or password")



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import time
import xmlrpc.client

from ...optconf import Option

from . import BaseClient
from . import WithCustoms
from . import NoSuchTorrentError
from . import hash_or_torrent
from . import check_torrent_accessible
from . import build_files


# =====
_XMLRPC_UNKNOWN_HASH = -501


# =====
def _catch_unknown_torrent(method):
    def wrap(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except xmlrpc.client.Fault as err:
            if err.faultCode == _XMLRPC_UNKNOWN_HASH:
                raise NoSuchTorrentError("Unknown torrent hash")
            raise
    return wrap


class Plugin(BaseClient, WithCustoms):
    # API description: http://code.google.com/p/gi-torrent/wiki/rTorrent_XMLRPC_reference

    def __init__(self, url, load_retries, retries_sleep, xmlrpc_size_limit, **kwargs):  # pylint:disable=super-init-not-called
        self._init_bases(**kwargs)

        self._load_retries = load_retries
        self._retries_sleep = retries_sleep
        self._server = xmlrpc.client.ServerProxy(url)
        self._server.set_xmlrpc_size_limit(xmlrpc_size_limit)

    @classmethod
    def get_name(cls):
        return "rtorrent"

    @classmethod
    def get_options(cls):
        return cls._get_merged_options({
            "url":               Option(default="http://localhost/RPC2", help="XMLRPC mountpoint"),
            "load_retries":      Option(default=10, help="The number of retries to load the torrent"),
            "retries_sleep":     Option(default=1.0, help="Sleep interval between the load retries"),
            "xmlrpc_size_limit": Option(default=67108863, help="Max XMLRPC data size"),
        })

    # ===

    @hash_or_torrent
    @_catch_unknown_torrent
    def remove_torrent(self, torrent_hash):
        self._server.d.erase(torrent_hash)

    @check_torrent_accessible
    def load_torrent(self, torrent, prefix=None):
        torrent_hash = torrent.get_hash()
        # XXX: https://github.com/rakshasa/rtorrent/issues/22
        # All load_* calls re asynchronous, so we need to wait until the load of torrent files is complete.
        self._server.load_raw(torrent.get_data())
        retries = self._load_retries
        while True:
            try:
                assert self._server.d.get_hash(torrent_hash).lower() == torrent_hash
                break
            except xmlrpc.client.Fault as err:
                if err.faultCode != _XMLRPC_UNKNOWN_HASH:
                    raise
                if retries == 0:
                    raise RuntimeError("Timed out torrent uploads after {} seconds".format(
                        self._load_retries * self._retries_sleep))
                retries -= 1
                time.sleep(self._retries_sleep)

        if prefix is not None:
            self._server.d.set_directory(torrent_hash, prefix)
        self._server.d.start(torrent_hash)

    @hash_or_torrent
    def has_torrent(self, torrent_hash):
        try:
            assert self._server.d.get_hash(torrent_hash).lower() == torrent_hash
            return True
        except xmlrpc.client.Fault as err:
            if err.faultCode != _XMLRPC_UNKNOWN_HASH:
                raise
        return False

    def get_hashes(self):
        return list(map(str.lower, self._server.download_list()))

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_torrent_path(self, torrent_hash):
        return self._server.d.get_loaded_file(torrent_hash)

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_data_prefix(self, torrent_hash):
        mc = xmlrpc.client.MultiCall(self._server)
        mc.d.get_directory(torrent_hash)
        mc.d.is_multi_file(torrent_hash)
        (path, is_multi_file) = mc()
        if is_multi_file:
            path = os.path.dirname(os.path.normpath(path))
        return path

    def get_data_prefix_default(self):
        return self._server.get_directory()

    # ===

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_full_path(self, torrent_hash):
        return self._server.d.get_base_path(torrent_hash)

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_file_name(self, torrent_hash):
        return self._server.d.get_name(torrent_hash)

    @hash_or_torrent
    @_catch_unknown_torrent
    def is_single_file(self, torrent_hash):
        return (not self._server.d.is_multi_file(torrent_hash))

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_files(self, torrent_hash, on_fs=False):
        mc = xmlrpc.client.MultiCall(self._server)
        mc.d.get_base_path(torrent_hash)
        mc.d.get_base_filename(torrent_hash)
        mc.d.is_multi_file(torrent_hash)
        mc.d.get_size_files(torrent_hash)
        mc.f.get_size_bytes(torrent_hash, 0)
        (base_path, base_file_name, is_multi_file, count, first_file_size) = tuple(mc())
        base = (base_path if on_fs else base_file_name)

        if not is_multi_file:
            return {base: {"size": first_file_size}}

        mc = xmlrpc.client.MultiCall(self._server)
        for index in range(count):
            mc.f.get_path(torrent_hash, index)
            mc.f.get_size_bytes(torrent_hash, index)
        flist = tuple(mc())
        flist = tuple(zip(flist[::2], flist[1::2]))

        files = build_files(base, flist)
        files.update({base: None})
        return files

    # ===

    @classmethod
    def get_custom_keys(cls):
        return ("c1", "c2", "c3", "c4", "c5")

    @hash_or_torrent
    @_catch_unknown_torrent
    def set_customs(self, torrent_hash, customs):
        assert len(customs) != 0, "Empty customs dict"
        mc = xmlrpc.client.MultiCall(self._server)
        for (key, value) in customs.items():
            getattr(mc.d, "set_custom{}".format(key[1:]))(torrent_hash, value)
        mc()

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_customs(self, torrent_hash, keys):
        assert len(keys) != 0, "Empty customs keys list"
        keys = tuple(set(keys))
        mc = xmlrpc.client.MultiCall(self._server)
        for key in keys:
            getattr(mc.d, "get_custom{}".format(key[1:]))(torrent_hash)
        return dict(zip(keys, tuple(mc())))



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse
import re
import operator

from datetime import datetime

from dateutil.relativedelta import relativedelta

from ... import tfile

from . import BaseFetcher
from . import WithLogin
from . import WithTime


# =====
def _encode(arg):
    return arg.encode("utf-8")


def _decode(arg):
    return arg.decode("utf-8")


class Plugin(BaseFetcher, WithLogin, WithTime):
    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)
        self._comment_regexp = re.compile(r"http://tr\.anidub\.com/\?newsid=(\d+)")
        self._tzinfo = None
        self._last_data = None

    @classmethod
    def get_name(cls):
        return "tr.anidub.com"

    @classmethod
    def get_version(cls):
        return 0

    @classmethod
    def get_fingerprint(cls):
        return {
            "url":      "http://tr.anidub.com",
            "encoding": "utf-8",
            "text":     "<link rel=\"search\" type=\"application/opensearchdescription+xml\""
                        " href=\"http://tr.anidub.com/engine/opensearch.php\" title=\"AniDUB Tracker\" />",
        }

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    # ===

    def is_torrent_changed(self, torrent):
        self._last_data = None
        self._assert_match(torrent)
        page = _decode(self._read_url(torrent.get_comment()))
        if torrent.get_mtime() < self._get_upload_time(page):
            candidate = self._get_candidate(page, torrent)
            if torrent.get_hash() != candidate.get_hash():
                self._last_data = candidate.get_data()
                return True
        return False

    def _get_upload_time(self, page):
        date_match = re.search(r"<li><b>Дата:</b> ([^,\s]+, \d\d:\d\d)</li>", page)
        self._assert_logic(date_match is not None, "Upload date not found")
        date = date_match.group(1)

        now = datetime.now(self._tzinfo)
        day_template = "{date.day:02d}-{date.month:02d}-{date.year}"
        if "Сегодня" in date:
            date = date.replace("Сегодня", day_template.format(date=now))
        if "Вчера" in date:
            yesterday = now - relativedelta(days=1)
            date = date.replace("Вчера", day_template.format(date=yesterday))
        date += " " + datetime.now(self._tzinfo).strftime("%z")

        upload_time = int(datetime.strptime(date, "%d-%m-%Y, %H:%M %z").strftime("%s"))
        return upload_time

    def _get_candidate(self, page, torrent):
        downloads = set(map(int, re.findall(r"<a href=\"/engine/download.php\?id=(\d+)\" class=\" \">", page)))
        candidates = {}
        for download_id in downloads:
            data = self._read_url(
                url="http://tr.anidub.com/engine/download.php?id={}".format(download_id),
                headers={"Referer": torrent.get_comment()},
            )
            self._assert_valid_data(data)
            candidate = tfile.Torrent(data=data)

            name = candidate.get_name()
            candidates.setdefault(name, [])
            candidates[name].append((candidate, str(download_id)))

        name = torrent.get_name()
        self._assert_logic(name in candidates, "Can't find torrent named '{}' in downloads".format(name))
        self._assert_logic(len(candidates[name]) == 1, "Too many variants to download: {}".format(
                                                       ", ".join(map(operator.itemgetter(1), candidates[name]))))
        return candidates[name][0][0]

    def fetch_new_data(self, torrent):
        return self._last_data

    # ===

    def login(self):
        self._assert_auth(self._user is not None, "Required user for site")
        self._assert_auth(self._passwd is not None, "Required password for site")
        post = {
            "login_name":      _encode(self._user),
            "login_password":  _encode(self._passwd),
            "login":           b"submit",
        }
        post_data = _encode(urllib.parse.urlencode(post))
        page = _decode(self._read_url("http://tr.anidub.com/", data=post_data))
        profile = "<li><a href=\"http://tr.anidub.com/user/{}/\">Мой профиль</a></li>".format(self._user)
        self._assert_auth(profile in page, "Invalid user or password")
        self._tzinfo = self._select_tzinfo("Etc/GMT+4")



import os


# =====
def as_string_or_none(value):
    if value is None:
        return None
    return str(value)


def as_string_list(values):
    if isinstance(values, str):
        values = [values]
    return list(map(str, values))


def as_key_value(values):
    if isinstance(values, dict):
        return values
    return dict(
        tuple(map(str.strip, (item.split("=", 1) + [""])[:2]))
        for item in as_string_list(values)
        if len(item.split("=", 1)[0].strip()) != 0
    )


def as_path(value):
    return os.path.abspath(os.path.expanduser(value))


def as_paths_list(values):
    if isinstance(values, str):
        values = [values]
    return list(map(as_path, values))


def as_path_or_none(value):
    if value is None:
        return None
    return as_path(value)


def as_8int_or_none(value):
    if value is None:
        return None
    return int(str(value), 8)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import json

from . import tcollection


# =====
def get_cache(cache_path, force_rebuild, client, torrents_dir_path, name_filter, log):
    cache = _read(cache_path, force_rebuild, log)
    if _update(cache, client, torrents_dir_path, name_filter, log):
        _write(cache, cache_path, log)
    return cache


# =====
def _read(path, force_rebuild, log):
    fallback = {
        "version":  0,
        "torrents": {},
    }
    if force_rebuild or not os.path.exists(path):
        return fallback

    log.info("Reading the cache from {cyan}%s{reset} ...", (path,))
    with open(path) as cache_file:
        try:
            cache = json.loads(cache_file.read())
            if cache["version"] != fallback["version"]:
                return fallback
            else:
                return cache
        except (KeyError, ValueError):
            log.error("The cache was damaged - ingored: {red}%s{reset}", (path,))
            return fallback


def _write(cache, path, log):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    log.info("Writing the cache to {cyan}%s{reset} ...", (path,))
    with open(path, "w") as cache_file:
        cache_file.write(json.dumps(cache))


def _update(cache, client, path, name_filter, log):
    log.info("Fetching all hashes from client ...")
    hashes = client.get_hashes()

    log.info("Validating the cache ...")

    # --- Old ---
    to_remove = tuple(sorted(set(cache["torrents"]).difference(hashes)))
    if len(to_remove) != 0:
        for torrent_hash in to_remove:
            cache["torrents"].pop(torrent_hash)
        log.info("Removed {magenta}%d{reset} obsolete hashes from cache", (len(to_remove),))

    # --- New ---
    to_add = tuple(sorted(set(hashes).difference(cache["torrents"])))
    added = 0
    if len(to_add) != 0:
        torrents = tcollection.load_from_dir(path, name_filter, log)
        torrents = tcollection.by_hash(torrents)

        if not log.isatty():
            log.info("Adding files for the new {yellow}%d{reset} hashes ...", (len(to_add),))
        for torrent_hash in to_add:
            torrent = torrents.get(torrent_hash)
            if torrent is not None:
                name = os.path.basename(torrent.get_path())
                if log.isatty():
                    log.info("Adding files for {yellow}%s{reset} ...", (name,), one_line=True)
                cache["torrents"][torrent_hash] = {
                    # "name":      name,
                    # "is_single": torrent.is_single_file(),
                    "files":     torrent.get_files(),
                    "prefix":    client.get_data_prefix(torrent),
                }
                added += 1
            else:
                log.error("Not cached - missing torrent for: {red}%s{reset} -- %s",
                          (torrent_hash, client.get_file_name(torrent_hash)))
        if added != 0:
            log.info("Added {magenta}%d{reset} new hashes from client", (added,))

    return bool(len(to_remove) or added)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse
import re

from datetime import datetime

from . import BaseFetcher
from . import WithLogin
from . import WithCaptcha
from . import WithTime


# =====
def _encode(arg):
    return arg.encode("cp1251")


def _decode(arg):
    return arg.decode("cp1251")


class Plugin(BaseFetcher, WithLogin, WithCaptcha, WithTime):
    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)
        self._comment_regexp = re.compile(r"http://pornolab\.net/forum/viewtopic\.php\?t=(\d+)")
        self._tzinfo = None

    @classmethod
    def get_name(cls):
        return "pornolab.net"

    @classmethod
    def get_version(cls):
        return 1

    @classmethod
    def get_fingerprint(cls):
        return {
            "url":      "http://pornolab.net/forum/index.php",
            "encoding": "cp1251",
            "text":     "<link rel=\"search\" type=\"application/opensearchdescription+xml\" title=\"Поиск на Pornolab.net\""
                        " href=\"http://static.pornolab.net/opensearch.xml\">",
        }

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    # ===

    def is_torrent_changed(self, torrent):
        self._assert_match(torrent)
        page = _decode(self._read_url(torrent.get_comment()))
        return (torrent.get_mtime() < self._get_upload_time(page))

    def _get_upload_time(self, page):
        date_match = re.search(r"<span title=\"Зарегистрирован\">\[ (\d\d-([а-яА-Я]{3})-\d\d \d\d:\d\d) \]</span>", page)
        self._assert_logic(date_match is not None, "Upload date not found")
        date = date_match.group(1)
        date_month = date_match.group(2)

        date = date.replace(date_month, {
            month: str(number) for (number, month) in enumerate(
                ("Янв", "Фев", "Мар", "Апр", "Май", "Июн",
                 "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"), 1
            )
        }[date_month])  # TODO: Send shitbeam to datetime authors
        date += " " + datetime.now(self._tzinfo).strftime("%z")

        upload_time = int(datetime.strptime(date, "%d-%m-%y %H:%M %z").strftime("%s"))
        return upload_time

    def fetch_new_data(self, torrent):
        self._assert_match(torrent)
        topic_id = self._comment_regexp.match(torrent.get_comment()).group(1)
        data = self._read_url("http://pornolab.net/forum/dl.php?t={}".format(topic_id), data=b"")
        self._assert_valid_data(data)
        return data

    # ===

    def login(self):
        self._assert_auth(self._user is not None, "Required user for site")
        self._assert_auth(self._passwd is not None, "Required passwd for site")

        post = {
            "login_username": _encode(self._user),
            "login_password": _encode(self._passwd),
            "login":          b"\xc2\xf5\xee\xe4",
        }
        page = self._read_login(post)

        cap_static_regexp = re.compile(r"\"(http://static\.pornolab\.net/captcha/[^\"]+)\"")
        cap_static_match = cap_static_regexp.search(page)
        if cap_static_match is not None:
            cap_sid_match = re.search(r"name=\"cap_sid\" value=\"([a-zA-Z0-9]+)\"", page)
            cap_code_match = re.search(r"name=\"(cap_code_[a-zA-Z0-9]+)\"", page)
            self._assert_auth(cap_sid_match is not None, "Unknown cap_sid")
            self._assert_auth(cap_code_match is not None, "Unknown cap_code")

            post[cap_code_match.group(1)] = self._captcha_decoder(cap_static_match.group(1))
            post["cap_sid"] = cap_sid_match.group(1)
            page = self._read_login(post)
            self._assert_auth(cap_static_regexp.search(page) is None, "Invalid user, password or captcha")

        self._tzinfo = self._get_tzinfo(page)

    def _read_login(self, post):
        return _decode(self._read_url(
            url="http://pornolab.net/forum/login.php",
            data=_encode(urllib.parse.urlencode(post)),
        ))

    def _get_tzinfo(self, page):
        timezone_match = re.search(r"<p>Часовой пояс: <span class='tz_time'>(GMT [+-] \d{1,2})</span></p>", page)
        timezone = (timezone_match and "Etc/" + timezone_match.group(1).replace(" ", ""))
        return self._select_tzinfo(timezone)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os

from . import BaseClient
from . import NoSuchTorrentError
from . import hash_or_torrent
from . import check_torrent_accessible
from . import build_files

try:
    import dbus  # pylint: disable=import-error
except ImportError:
    dbus = None


# =====
class Plugin(BaseClient):
    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)

        if dbus is None:
            raise RuntimeError("Required module dbus")

        self._bus = dbus.SessionBus()
        self._core = self._bus.get_object("org.ktorrent.ktorrent", "/core")
        self._settings = self._bus.get_object("org.ktorrent.ktorrent", "/settings")

        if self._settings.useSaveDir():
            raise RuntimeError("Turn off the path by default in the settings of KTorrent")

    @classmethod
    def get_name(cls):
        return "ktorrent"

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    # ===

    @hash_or_torrent
    def remove_torrent(self, torrent_hash):
        self._get_torrent_obj(torrent_hash)  # XXX: raise NoSuchTorrentError if torrent does not exist
        self._core.remove(torrent_hash, False)

    @check_torrent_accessible
    def load_torrent(self, torrent, prefix=None):
        if prefix is not None:
            self._settings.setLastSaveDir(prefix)
        self._core.loadSilently(torrent.get_path(), "")

    @hash_or_torrent
    def has_torrent(self, torrent_hash):
        try:
            self._get_torrent_obj(torrent_hash)
            return True
        except NoSuchTorrentError:
            return False

    def get_hashes(self):
        return list(map(str.lower, self._core.torrents()))

    @hash_or_torrent
    def get_torrent_path(self, torrent_hash):
        raise RuntimeError("KTorrent can not return a path of the torrent file")

    @hash_or_torrent
    def get_data_prefix(self, torrent_hash):
        return str(self._get_torrent_obj(torrent_hash).dataDir())

    def get_data_prefix_default(self):
        raise RuntimeError("KTorrent can not return a default data path")

    # ===

    @hash_or_torrent
    def get_full_path(self, torrent_hash):
        return str(self._get_torrent_obj(torrent_hash).pathOnDisk())

    @hash_or_torrent
    def get_file_name(self, torrent_hash):
        return str(self._get_torrent_obj(torrent_hash).name())

    @hash_or_torrent
    def is_single_file(self, torrent_hash):
        return (self._get_torrent_obj(torrent_hash).numFiles() == 0)

    @hash_or_torrent
    def get_files(self, torrent_hash, on_fs=False):
        torrent_obj = self._get_torrent_obj(torrent_hash)
        prefix = (str(torrent_obj.pathOnDisk()) if on_fs else "")
        count = torrent_obj.numFiles()
        name = str(torrent_obj.name())

        if count == 0:  # Single file
            flist = [(name, int(torrent_obj.totalSize()))]
        else:
            flist = [
                (
                    os.path.join(name, str(torrent_obj.filePath(dbus.UInt32(index)))),
                    int(torrent_obj.fileSize(dbus.UInt32(index))),
                )
                for index in range(count)
            ]
        return build_files(prefix, flist)

    # ===

    def _get_torrent_obj(self, torrent_hash):
        if torrent_hash not in self.get_hashes():
            raise NoSuchTorrentError("Unknown torrent hash")
        try:
            torrent_obj = self._bus.get_object("org.ktorrent.ktorrent", "/torrent/" + torrent_hash)
            assert str(torrent_obj.infoHash()) == torrent_hash
            return torrent_obj
        except dbus.exceptions.DBusException as err:
            if err.get_dbus_name() == "org.freedesktop.DBus.Error.UnknownObject":
                raise NoSuchTorrentError("Unknown torrent hash")
            raise



#!/usr/bin/env python3
"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import setuptools


# =====
if __name__ == "__main__":
    setuptools.setup(
        name="emonoda",
        version="1.9.22",
        url="https://github.com/mdevaev/emonoda",
        license="GPLv3",
        author="Devaev Maxim",
        author_email="mdevaev@gmail.com",
        description="A set of tools to organize and manage your torrents",
        platforms="any",

        packages=[
            "emonoda",
            "emonoda.optconf",
            "emonoda.apps",
            "emonoda.apps.hooks",
            "emonoda.helpers",
            "emonoda.plugins",
            "emonoda.plugins.clients",
            "emonoda.plugins.fetchers",
            "emonoda.plugins.confetti",
            "emonoda.thirdparty",
        ],

        package_data={
            "emonoda.plugins.confetti": ["templates/*.mako"],
        },

        entry_points={
            "console_scripts": [
                "emdiff = emonoda.apps.emdiff:main",
                "emfetch = emonoda.apps.emfetch:main",
                "emfile = emonoda.apps.emfile:main",
                "emload = emonoda.apps.emload:main",
                "emfind = emonoda.apps.emfind:main",
                "emrm = emonoda.apps.emrm:main",
                "emhook-manage-trackers = emonoda.apps.hooks.manage_trackers:main",
            ],
        },

        install_requires=[
            "chardet",
            "pyyaml",
            "colorama",
            "pygments",
            "pytz",
            "python-dateutil",
            "Mako",
        ],

        classifiers=[
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Development Status :: 5 - Production/Stable",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: Utilities",
            "Operating System :: OS Independent",
        ],
    )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import socket
import urllib.request
import urllib.parse
import http.client
import time

from . import sockshandler


# =====
def build_opener(proxy_url=None, cookie_jar=None):
    handlers = []

    if proxy_url is not None:
        scheme = (urllib.parse.urlparse(proxy_url).scheme or "").lower()
        if scheme == "http":
            proxies = dict.fromkeys(("http", "https"), proxy_url)
            handlers.append(urllib.request.ProxyHandler(proxies))
        elif scheme in ("socks4", "socks5"):
            handlers.append(sockshandler.SocksHandler(proxy_url=proxy_url))
        else:
            raise RuntimeError("Invalid proxy protocol: {}".format(scheme))

    if cookie_jar is not None:
        handlers.append(urllib.request.HTTPCookieProcessor(cookie_jar))

    return urllib.request.build_opener(*handlers)


def read_url(
    opener,
    url,
    data=None,
    headers=None,
    timeout=10,
    retries=10,
    retries_sleep=1,
    retry_codes=(500, 502, 503),
    retry_timeout=True,
):
    while True:
        try:
            request = urllib.request.Request(url, data, (headers or {}))
            return opener.open(request, timeout=timeout).read()
        except socket.timeout as err:
            if retries == 0 or not retry_timeout:
                raise
        except urllib.error.HTTPError as err:
            if retries == 0 or err.code not in retry_codes:
                raise
        except urllib.error.URLError as err:
            if "timed out" in str(err.reason):
                if retries == 0 or not retry_timeout:
                    raise
            else:
                raise
        except (http.client.IncompleteRead, http.client.BadStatusLine, ConnectionResetError) as err:
            if retries == 0:
                raise
        time.sleep(retries_sleep)
        retries -= 1



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import contextlib
import argparse

import pygments
import pygments.lexers.data
import pygments.formatters

from ..optconf import make_config
from ..optconf import Section
from ..optconf import Option

from ..optconf import build_raw_from_options
from ..optconf.dumper import make_config_dump
from ..optconf.loader import load_file as load_yaml_file
from ..optconf.converters import (
    as_string_or_none,
    as_string_list,
    as_key_value,
    as_path,
    as_paths_list,
    as_path_or_none,
    as_8int_or_none,
)

from ..plugins import get_client_class
from ..plugins import get_fetcher_class
from ..plugins import get_confetti_class

from ..plugins.fetchers import WithLogin as F_WithLogin
from ..plugins.fetchers import WithCaptcha as F_WithCaptcha
from ..plugins.clients import WithCustoms as C_WithCustoms

from .. import cli


# =====
def init():
    args_parser = argparse.ArgumentParser(add_help=False)
    args_parser.add_argument("-c", "--config", dest="config_file_path", default="~/.config/emonoda.yaml", metavar="<file>")
    args_parser.add_argument("-o", "--set-options", dest="set_options", default=[], nargs="+")
    args_parser.add_argument("-m", "--dump-config", dest="dump_config", action="store_true")
    (options, remaining) = args_parser.parse_known_args(sys.argv)

    options.config_file_path = os.path.expanduser(options.config_file_path)
    if os.path.exists(options.config_file_path):
        raw_config = load_yaml_file(options.config_file_path)
    else:
        raw_config = {}
    _merge_dicts(raw_config, build_raw_from_options(options.set_options))
    scheme = _get_config_scheme()
    config = make_config(raw_config, scheme)

    if config.core.client is not None:
        client_scheme = get_client_class(config.core.client).get_options()
        scheme["client"] = client_scheme

    for fetcher_name in raw_config.get("fetchers", []):
        fetcher_scheme = get_fetcher_class(fetcher_name).get_options()
        scheme.setdefault("fetchers", {})
        scheme["fetchers"][fetcher_name] = fetcher_scheme

    for confetti_name in raw_config.get("confetti", []):
        confetti_scheme = get_confetti_class(confetti_name).get_options()
        scheme.setdefault("confetti", {})
        scheme["confetti"][confetti_name] = confetti_scheme

    config = make_config(raw_config, scheme)

    if options.dump_config:
        dump = make_config_dump(config)
        if sys.stdout.isatty():
            dump = pygments.highlight(
                dump,
                pygments.lexers.data.YamlLexer(),
                pygments.formatters.TerminalFormatter(bg="dark"),
            )
        print(dump)
        sys.exit(0)

    config.setdefault("client", Section())
    config.setdefault("fetchers", Section())
    config.setdefault("confetti", Section())

    return (args_parser, remaining, config)


def _merge_dicts(dest, src, path=None):
    if path is None:
        path = []
    for key in src:
        if key in dest:
            if isinstance(dest[key], dict) and isinstance(src[key], dict):
                _merge_dicts(dest[key], src[key], list(path) + [str(key)])
                continue
        dest[key] = src[key]


# =====
@contextlib.contextmanager
def get_configured_log(config, quiet, output):
    log = cli.Log(config.core.use_colors, config.core.force_colors, quiet, output)
    try:
        yield log
    finally:
        log.finish()


def get_configured_client(config, required, with_customs, log):
    name = config.core.client
    if name is not None:
        log.info("Enabling the client {blue}%s{reset} ...", (name,), one_line=True)
        try:
            client = get_client_class(name)(**config.client)
            if with_customs and C_WithCustoms not in client.get_bases():
                raise RuntimeError("Your client does not support customs")
        except Exception as err:
            log.error("Can't init client {red}%s{reset}: {red}%s{reset}(%s)", (name, type(err).__name__, err))
            raise
        log.info("Client {blue}%s{reset} is {green}ready{reset}", (name,))
        return client
    elif required:
        raise RuntimeError("No configured client found")
    else:
        return None


def get_configured_fetchers(config, captcha_decoder, only, exclude, log):
    to_init = set(config.fetchers).difference(exclude)
    if len(only) != 0:
        to_init = to_init.intersection(only)

    if len(to_init) == 0:
        raise RuntimeError("No fetchers to init")

    fetchers = []
    for fetcher_name in sorted(to_init):
        log.info("Enabling the fetcher {blue}%s{reset} ...", (fetcher_name,), one_line=True)

        fetcher_class = get_fetcher_class(fetcher_name)
        fetcher_kwargs = dict(config.fetchers[fetcher_name])
        if F_WithCaptcha in fetcher_class.get_bases():
            fetcher_kwargs["captcha_decoder"] = captcha_decoder
        fetcher = fetcher_class(**fetcher_kwargs)

        try:
            log.info("Enabling the fetcher {blue}%s{reset}: {yellow}testing{reset} ...", (fetcher_name,), one_line=True)
            fetcher.test()
            if F_WithLogin in fetcher_class.get_bases():
                log.info("Enabling the fetcher {blue}%s{reset}: {yellow}logging in{reset} ...", (fetcher_name,), one_line=True)
                fetcher.login()
            log.info("Fetcher {blue}%s{reset} is {green}ready{reset}", (fetcher_name,))
        except Exception as err:
            log.error("Can't init fetcher {red}%s{reset}: {red}%s{reset}(%s)", (fetcher_name, type(err).__name__, err))
            raise

        fetchers.append(fetcher)

    return fetchers


def get_configured_confetti(config, log):
    confetti = []
    for confetti_name in sorted(config.confetti):
        log.info("Enabling the confetti {blue}%s{reset} ...", (confetti_name,), one_line=True)
        confetti_class = get_confetti_class(confetti_name)
        confetti_kwargs = dict(config.confetti[confetti_name])
        try:
            confetti.append(confetti_class(**confetti_kwargs))
            log.info("Confetti {blue}%s{reset} is {green}ready{reset}", (confetti_name,))
        except Exception as err:
            log.error("Can't init confetti {red}%s{reset}: {red}%s{reset}(%s)", (confetti_name, type(err).__name__, err))
            raise
    return confetti


# =====
def _get_config_scheme():
    return {
        "core": {
            "client":        Option(default=None, type=as_string_or_none, help="The name of plugin for torrent client"),
            "torrents_dir":  Option(default=".", type=as_path, help="Path to directory with torrent files"),
            "data_root_dir": Option(default="~/Downloads", type=as_path, help="Path to root directory with data of torrents"),
            "another_data_root_dirs": Option(default=[], type=as_paths_list, help="Paths to another data directories"),
            "use_colors":    Option(default=True, help="Enable colored output"),
            "force_colors":  Option(default=False, help="Always use the coloring"),
        },

        "emfetch": {
            "name_filter":   Option(default="*.torrent", type=str, help="Update only filtered torrent files"),
            "backup_dir":    Option(default=None, type=as_path_or_none, help="Backup old torrent files after update here"),
            "backup_suffix": Option(default=".%Y.%m.%d-%H:%M:%S.bak", help="Append this suffix to backuped file"),
            "save_customs":  Option(default=[], type=as_string_list, help="Save client custom fields after update if supports"),
            "set_customs":   Option(default={}, type=as_key_value, help="Set client custom fileds after update if supports"),
            "show_unknown":  Option(default=False, help="Show the torrents with unknown tracker in the log"),
            "show_passed":   Option(default=False, help="Show the torrents without changes"),
            "show_diff":     Option(default=True, help="Show diff between old and updated torrent files"),
        },

        "emload": {
            "torrent_mode": Option(default=None, type=as_8int_or_none, help="Change permissions of torrent file before load"),
            "mkdir_mode": Option(default=None, type=as_8int_or_none, help="Permission for new directories"),
            "set_customs":   Option(default={}, type=as_key_value, help="Set client custom fileds after update if supports")
        },

        "emfind": {
            "cache_file": Option(default=as_path("~/.cache/emfind.json"), type=as_path, help="Torrents cache"),
            "name_filter": Option(default="*.torrent", type=str, help="Cache only filtered torrent files"),
        },
    }



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import re
import hashlib
import base64
import urllib.parse
import itertools

import chardet

from .thirdparty.bcoding import bdecode as decode_data
from .thirdparty.bcoding import bencode as encode_struct


# =====
def is_valid_data(data):
    try:
        return isinstance(decode_data(data), dict)  # Must be a True
    except TypeError:
        return False


def is_hash(text):
    return (re.match(r"[\da-fA-F]{40}", text) is not None)


def get_difference(old, new):
    assert isinstance(old, (Torrent, dict))
    assert isinstance(new, (Torrent, dict))
    old = (old.get_files() if isinstance(old, Torrent) else old)
    new = (new.get_files() if isinstance(new, Torrent) else new)

    modified = set()
    type_modified = set()
    for path in set(old).intersection(set(new)):
        old_attrs = old[path]
        new_attrs = new[path]

        real = len(tuple(filter(None, (new_attrs, old_attrs))))
        if real == 0:
            continue
        elif real == 1:
            type_modified.add(path)
            continue

        if old_attrs["size"] != new_attrs["size"]:
            modified.add(path)

    return {
        "added":         set(new).difference(set(old)),
        "removed":       set(old).difference(set(new)),
        "modified":      modified,
        "type_modified": type_modified,
    }


# =====
class Torrent:
    def __init__(self, data=None, path=None):
        # https://wiki.theory.org/BitTorrentSpecification

        self._path = path
        self._data = None
        self._bencode = None
        self._hash = None
        self._scrape_hash = None

        if data is not None:
            self.load_from_data(data, path)
        elif path is not None:
            self.load_from_file(path)

    def load_from_file(self, path):
        with open(path, "rb") as torrent_file:
            return self.load_from_data(torrent_file.read(), path)

    def load_from_data(self, data, path=None):
        self._bencode = decode_data(data)
        self._path = path
        self._data = data
        self._hash = None
        self._scrape_hash = None
        return self

    # ===

    def get_path(self):
        return self._path

    def get_mtime(self):
        assert self._path is not None, "Can't get mtime for None-path"
        return int(os.stat(self._path).st_mtime)

    def get_data(self):
        return self._data

    def get_bencode(self):
        return self._bencode

    # ===

    def get_name(self):
        return self._decode(self._bencode["info"]["name"])

    def get_comment(self):
        return self._decode(self._bencode.get("comment"))

    def get_encoding(self):
        return self._bencode.get("encoding")

    def get_creation_date(self):
        return self._bencode.get("creation date")

    def get_created_by(self):
        return self._bencode.get("created by")

    def get_announce(self):
        return self._bencode.get("announce")

    def get_announce_list(self):
        return self._bencode.get("announce-list", [])

    def is_private(self):
        return bool(self._bencode["info"].get("private", 0))

    # ===

    def get_hash(self):
        if self._hash is None:
            self._hash = hashlib.sha1(encode_struct(self._bencode["info"])).hexdigest().lower()
        return self._hash

    def get_scrape_hash(self):
        if self._scrape_hash is None:
            scrape_hash = ""
            torrent_hash = self.get_hash()
            for index in range(0, len(torrent_hash), 2):
                scrape_hash += "%{}".format(torrent_hash[index:index + 2])
            self._scrape_hash = scrape_hash
        return self._scrape_hash

    def make_magnet(self, extras=()):
        # http://stackoverflow.com/questions/12479570/given-a-torrent-file-how-do-i-generate-a-magnet-link-in-python
        info_sha1 = hashlib.sha1(encode_struct(self._bencode["info"]))
        info_digest = info_sha1.digest()
        b32_hash = base64.b32encode(info_digest)

        magnet = "magnet:?xt={}".format(urllib.parse.quote_plus("urn:btih:{}".format(b32_hash)))
        if "name" in extras:
            magnet += "&dn={}".format(urllib.parse.quote_plus(self.get_name()))
        if "trackers" in extras:
            announces = tuple(filter(None, [self.get_announce()] + self.get_announce_list()))
            for announce in set(itertools.chain.from_iterable(announces)):
                magnet += "&tr={}".format(urllib.parse.quote_plus(announce))
        if "size" in extras:
            magnet += "&xl={}".format(self.get_size())
        return magnet

    # ===

    def get_size(self):
        if self.is_single_file():
            return self._bencode["info"]["length"]
        else:
            size = 0
            for fstruct in self._bencode["info"]["files"]:
                size += fstruct["length"]
            return size

    def is_single_file(self):
        return ("files" not in self._bencode["info"])

    def get_files(self, prefix=""):
        make_file_attrs = (lambda fstruct: {"size": fstruct["length"]})
        base = os.path.join(prefix, self.get_name())
        if self.is_single_file():
            return {base: make_file_attrs(self._bencode["info"])}
        else:
            files = {base: None}
            for fstruct in self._bencode["info"]["files"]:
                name = None
                for index in range(len(fstruct["path"])):
                    name = os.path.join(base, os.path.sep.join(map(self._decode, fstruct["path"][0:index + 1])))
                    files[name] = None
                assert name is not None
                files[name] = make_file_attrs(fstruct)
            return files

    # ===

    def _decode(self, value):
        if isinstance(value, bytes):
            for encoding in (self._bencode.get("encoding", "utf-8"), "cp1251"):
                try:
                    return value.decode(encoding)
                except UnicodeDecodeError:
                    pass
            encoding = chardet.detect(value)["encoding"]
            assert encoding is not None, "Can't determine encoding for bytes string: '{}'".format(repr(value))
            return value.decode(encoding)
        else:
            return value



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import math
import datetime


# =====
_UNITS = tuple(zip(
    ("bytes", "kB", "MB", "GB", "TB", "PB"),
    (0,       0,    1,    2,    2,    2),
))


# =====
def format_size(size):
    if size > 1:
        exponent = min(int(math.log(size, 1024)), len(_UNITS) - 1)
        quotient = float(size) / 1024 ** exponent
        (unit, decimals) = _UNITS[exponent]
        result = ("{:.%sf} {}" % (decimals)).format(quotient, unit)
    elif size == 0:
        result = "0 bytes"
    elif size == 1:
        result = "1 byte"
    else:
        ValueError("size must be >= 0")
    return result


def format_progress(value, limit):
    return (("%%%dd/" % (len(str(limit)))) + "%d") % (value, limit)


def format_now(text):
    return datetime.datetime.now().strftime(text)


def format_torrents_diff(diff, prefix):
    lines = []
    placeholders = ()
    for (sign, color, items) in (
        ("+", "green",  diff["added"]),
        ("-", "red",    diff["removed"]),
        ("~", "cyan",   diff["modified"]),
        ("?", "yellow", diff["type_modified"]),
    ):
        for item in sorted(items):
            lines.append("%s{" + color + "}%s{reset} %s")
            placeholders += (prefix, sign, item)
    return ("\n".join(lines), placeholders)


# =====
def make_fan():
    fan = 0
    while True:
        if fan < 3:
            fan += 1
        else:
            fan = 0
        yield "/-\\|"[fan]
